from flask import Flask, request, redirect, url_for, abort, g
from flask_login import LoginManager
from pathlib import Path
from http import HTTPStatus
import os

from . import user
from . extensions import db, bcrypt, mail, migrate, installed_apps

installed_apps.append(user)

def create_app():
    app_directory = os.path.dirname(os.path.abspath(__file__))
    instance_folder = os.path.join(app_directory, 'instance')

    app = Flask(__name__, instance_path=instance_folder)
    app.config.from_pyfile('config.py', silent=True)

    instance_path = Path(app.instance_path)
    parent_directory = Path(instance_path.parent)
    if not parent_directory.is_dir():
        parent_directory.mkdir()
    
    if not instance_path.is_dir():
        instance_path.mkdir()
    
    
    login_manager = LoginManager()
    login_manager.init_app(app)

    @login_manager.user_loader
    def load_user(user_id):
        return user.User.query.get(user_id)

    @login_manager.unauthorized_handler
    def unauthorized():
        if request.blueprint == 'api':
            abort(HTTPStatus.UNAUTHORIZED)
        return redirect(url_for('user.login'))
    
    
    # Register Blueprints
    menu_list = []
    for module in installed_apps:
        app.register_blueprint(getattr(module, "bp"))
        if hasattr(module, "menu_label"):
            menu_list.append(getattr(module, "menu_label"))

    app.config['MENUS'] = menu_list

    # Initialize the database
    bcrypt.init_app(app)
    mail.init_app(app)
    db.init_app(app)
    migrate.init_app(app=app, db=db)
    
    return app



from flask_sqlalchemy import SQLAlchemy
from flask_mail import Mail
from flask_bcrypt import Bcrypt
from flask_migrate import Migrate

db = SQLAlchemy()
installed_apps = []
mail = Mail()
bcrypt = Bcrypt()
migrate = Migrate()



from flask import Blueprint, render_template, request, redirect, url_for, flash
from flask_login import login_user, current_user, logout_user
from functools import wraps

from .models import User, Role, UserRole
from .forms import LoginForm, UserForm

from .. extensions import db


bp = Blueprint("user", __name__, template_folder="pages", url_prefix="/user")


def login_required(func):
    @wraps(func)
    def decorated_view(*args, **kwargs):
        if not current_user.is_authenticated:
            return redirect(url_for('user.login'))
        
        if not current_user.is_active():
            return redirect(url_for("user.inactive"))
                
        return func(*args, **kwargs)
    return decorated_view


def roles_accepted(roles=[]):
    def decorator(func):
        @wraps(func)
        def decorated_view(*args, **kwargs):
            if not any(role in roles for role in current_user.user_roles):
                flash("You are not allowed to access this area.", category="error")
                return redirect(url_for('main.home')) 
            return func(*args, **kwargs)
        return decorated_view
    return decorator  


@bp.route("/")
@login_required
def home():
    return "User profile"


@bp.route("/list")
@login_required
@roles_accepted(['user'])
def user_list():
    users = User.query.order_by('user_name').all()
    roles = Role.query.order_by('role_name').all()
    context = {
        "users": users,
        "roles": roles
    }
    return render_template("user/user_list.html", **context)


@bp.route("user_admin")
@login_required
@roles_accepted(['user'])
def user_admin():
    user_id = request.args.get('user_id')
    value = int(request.args.get("value"))
    
    user = User.query.get(user_id)
    if user.user_name != "admin":
        user.admin = value
        db.session.commit()
    else:
        flash("Cannot change super admin status", category="error")
    
    return redirect(url_for('user.user_list'))


@bp.route("user/active")
@login_required
@roles_accepted(['user'])
def user_active():
    user_id = request.args.get('user_id')
    value = int(request.args.get("value"))

    print(user_id, value)
    
    user = User.query.get(user_id)
    if user.user_name != "admin":
        user.active = value   
        db.session.commit()
    else:
        flash("Cannot change super admin status", category="error")
    
    return redirect(url_for('user.user_list'))


@bp.route("/login", methods=["POST", "GET"])
def login():
    if current_user.is_authenticated:
        return redirect("/")
    
    if request.method == "POST":
        form = LoginForm()
        form.post(request.form)
        
        if form.validate():
            user = User.query.filter_by(user_name=form.user_name).first()
            if user:
                if user.check_pass_word(form.pass_word):
                    login_user(user)
                    flash(f"Welcome {user.user_name}.", category="success")
                    return redirect("/")
                
            flash("Invalid username / password.", category="error")
    else:
        form = LoginForm()
    
    check_roles()
    
    context = {
        "form": form,
    }
    return render_template("user/login.html", **context)


@bp.route("/register", methods=["POST", "GET"])
def register():
    if request.method == "POST":
        form = UserForm()
        form.post(request.form)
        if form.validate():
            user = User(
                user_name=form.user_name,
                first_name=form.first_name,
                middle_name=form.middle_name,
                last_name=form.last_name,
                email=form.email
            )
            
            if form.user_name == "admin":
                user.active = True
                user.admin = True                
                
            user.set_pass_word(form.pass_word)
            
            db.session.add(user)
            db.session.commit()
            
            if form.user_name == "admin":
                if not Role.query.all():
                    check_roles()
                
                role = Role.query.filter_by(role_name="user").first()
                                    
                user_role = UserRole(
                    user_id=user.id,
                    role_id=role.id
                )
                
                db.session.add(user_role)
                db.session.commit()               
            
            login_user(user)
            flash(f"Welcome {user.user_name}.", category="success")
            return redirect(url_for("user.inactive"))
        
    else:
        form = UserForm()
    
    context = {
        "form": form
    }
    return render_template("user/register.html", **context)


@bp.route("/logout")
def logout():
    if current_user.is_authenticated:
        logout_user()
        flash("User logged out.", category="success")
        
    return redirect(url_for('user.login'))


@bp.route("/inactive")
def inactive():
    if current_user.is_active():
        return redirect("/")
    
    return render_template("user/inactive.html")


@bp.route("/add_role")
@login_required
@roles_accepted(['user'])
def add_role():
    user_id = int(request.args.get("user_id"))
    role_id = int(request.args.get("role_id"))
    
    user_role = UserRole(
        user_id=user_id,
        role_id=role_id
    )
    
    db.session.add(user_role)
    db.session.commit()
    
    return redirect(url_for('user.user_list'))
    
@bp.route("/remove_role")
@login_required
@roles_accepted(['user'])
def remove_role():
    user_id = int(request.args.get("user_id"))
    role_id = int(request.args.get("role_id"))
    
    user = User.query.get(user_id)
    role = Role.query.get(role_id)
    
    if user.user_name == 'admin' and role.role_name == 'user':
        flash("Cannot remove user role for super admin", category='error')
    else:
        user_role = UserRole.query.filter_by(
            user_id=user_id,
            role_id=role_id
        ).first()
        
        db.session.delete(user_role)
        db.session.commit()
    
    return redirect(url_for('user.user_list'))


def check_roles():
    from .. extensions import installed_apps
    role_names = ['user']
    for module in installed_apps:
        if hasattr(module,"menu_label"):
            role_names.append(getattr(module,"menu_label")[2])
                    
    for role_name in role_names:
        role = Role.query.filter_by(role_name=role_name).first()
        if not role:
            role = Role(role_name=role_name)
            db.session.add(role)
            db.session.commit()


import os
from werkzeug.security import generate_password_hash, check_password_hash
from flask import session, current_app

from ..extensions import db


class Role(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    role_name = db.Column(db.String())
    
    def __str__(self):
        return self.role_name
    

class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user_name = db.Column(db.String())
    pass_word = db.Column(db.String())
    first_name = db.Column(db.String())
    middle_name = db.Column(db.String())
    last_name = db.Column(db.String())
    email = db.Column(db.String())
    admin = db.Column(db.Boolean(), default=False)
    staff = db.Column(db.Boolean(), default=False)
    active = db.Column(db.Boolean(), default=False)
    salt = db.Column(db.String())
    
    def __str__(self):
        return self.first_name + " " + self.last_name
    
    def set_pass_word(self, pass_word):
        salt = os.urandom(16)
        salted_password = f"{salt}{pass_word}"
        self.salt = salt
        self.pass_word = generate_password_hash(salted_password)
        
    def check_pass_word(self, pass_word):
        salted_pass_word = f"{self.salt}{pass_word}"
        return check_password_hash(self.pass_word, salted_pass_word)
    
    def is_active(self):
        return self.active
    
    def get_id(self):
        return self.id
    
    def is_authenticated(self):
        if 'user_id' in session: return True
        
    @property
    def user_roles(self):
        return [user_role.role.role_name for user_role in self.roles]
    
    @property
    def menus(self):
        return current_app.config['MENUS']


class UserRole(db.Model):
    user_id = db.Column(db.Integer, db.ForeignKey('user.id'), primary_key=True)
    user = db.relationship('User', backref='roles', lazy=True)
    
    role_id = db.Column(db.Integer, db.ForeignKey('role.id'), primary_key=True)
    role = db.relationship('Role', backref='users', lazy=True)



from setuptools import setup, find_packages

setup(
    name='acas_auth',
    version='2.0.1',
    packages=find_packages(),
    install_requires=[
        "Flask==3.0.0",
        "Flask-Bcrypt==1.0.1",
        "Flask-Login==0.6.3",
        "Flask-Mail==0.9.1",
        "Flask-Migrate==4.0.5"
    ]
)



from dataclasses import dataclass
import re
from sqlalchemy import func

from .models import User


@dataclass
class UserForm:
    user_name: str = ""
    pass_word: str = ""
    confirm_pass_word: str = ""
    first_name: str = ""
    middle_name: str = ""
    last_name: str = ""
    email: str = ""
    
    errors = {}
    
    def post(self, request_form):
        self.user_name = request_form.get("user_name")
        self.pass_word = request_form.get("pass_word")
        self.confirm_pass_word = request_form.get("confirm_pass_word")
        self.first_name = request_form.get("first_name")
        self.middle_name = request_form.get("middle_name")
        self.last_name = request_form.get("last_name")
        self.email = request_form.get("email")
        
    def validate(self):
        self.errors = {}
        
        if not self.user_name: 
            self.errors["user_name"] = "Please type username."
        elif not self.is_valid_char(self.user_name):
            self.errors["user_name"] = "User alpha-numeric and underscore for username."
        else:
            user = User.query.filter(func.upper(User.user_name) == func.upper(self.user_name)).first()
            if user:
                self.errors["user_name"] = "Username is already taken. Please choose another one."

            
        if not self.pass_word:
            self.errors["pass_word"] = "Please type password."
        elif len(self.pass_word) < 6:
            self.errors["pass_word"] = "Password should be at least six (6) characters."
        
        if not self.confirm_pass_word:
            self.errors["confirm_pass_word"] = "Please re-type password."
        elif self.pass_word != self.confirm_pass_word:
            self.errors["confirm_pass_word"] = "Passwords does not match."
            
        if not self.first_name: self.errors["first_name"] = "Please type first name."
        if not self.last_name: self.errors["last_name"] = "Please type last name."
        
        if not self.errors:
            return True
        
    def is_valid_char(self, value):
        # Regular expression pattern to match valid Excel sheet names
        pattern = re.compile(r'^[a-zA-Z0-9_.]*([a-zA-Z0-9_.]+$)')
        return pattern.match(value) is not None


@dataclass
class LoginForm:
    user_name: str = ""
    pass_word: str = ""
    
    errors = {}
    
    def post(self, request_form):
        self.user_name = request_form.get("user_name") if request_form.get("user_name") else ""
        self.pass_word = request_form.get("pass_word") if request_form.get("pass_word") else ""

    def validate(self):
        self.errors = {}
        if not self.user_name: 
            self.errors["user_name"] = "Please type username."
        if not self.pass_word: self.errors["pass_word"] = "Please type password."
        if not self.errors: return True
        